from django.db import models
from simple_history.models import HistoricalRecords


class Historized(models.Model):
    history = HistoricalRecords(inherit=True)

    class Meta:
        abstract = True



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Dated(models.Model):
    created = models.DateTimeField(
        editable=False, auto_now_add=True, verbose_name=_('created')
    )
    updated = models.DateTimeField(
        editable=False, auto_now=True, verbose_name=_('updated')
    )

    class Meta:
        abstract = True



import logging
from typing import Optional, Type, Tuple

from django.core.exceptions import ValidationError
from django.db import models, IntegrityError

LOGGER = logging.getLogger(__name__)


def get_object_or_none(model: Type[models.Model], **search_keys) \
        -> Optional[models.Model]:
    assert issubclass(model, models.Model)
    try:
        obj = model.objects.get(**search_keys)
    except model.DoesNotExist:
        obj = None
    return obj


def validate_and_create_object(model: Type[models.Model], **kwargs) \
        -> models.Model:
    """
    :raises ValueError
    :return obj
    """
    assert issubclass(model, models.Model)
    obj = model(**kwargs)
    try:
        obj.full_clean()
        obj.save()
    except (ValidationError, IntegrityError) as exc:
        LOGGER.warning(
            'Create %s error: %r (kwargs=%r)', model.__name__, exc, kwargs)
        raise ValueError(str(exc))
    return obj


def validate_and_update_object(obj: models.Model, **kwargs) \
        -> Tuple[models.Model, bool]:
    """
    :raises ValueError
    :return obj, is_updated
    """
    assert isinstance(obj, models.Model)
    model = type(obj)

    updated_keys = {}
    for key, value in kwargs.items():
        old_value = getattr(obj, key)
        if old_value == value:
            continue
        setattr(obj, key, value)
        updated_keys[key] = old_value

    if updated_keys:
        try:
            obj.full_clean()
            obj.save()
        except (ValidationError, IntegrityError) as exc:
            for key, old_value in updated_keys.items():
                setattr(obj, key, old_value)
            LOGGER.warning(
                'Update %s error: %r (kwargs=%r)', model.__name__, exc, kwargs)
            raise ValueError(str(exc))
    return obj, bool(updated_keys)


def update_or_create_object(
        model: Type[models.Model],
        search_keys: Optional[dict] = None,
        **kwargs) \
        -> Tuple[models.Model, bool, bool]:
    """
    :raises ValueError
    :return obj, is_updated, is_created
    """
    obj = get_object_or_none(model, **search_keys) if search_keys else None
    if obj:
        is_created = False
        obj, is_updated = validate_and_update_object(obj, **kwargs)
    else:
        is_updated = False
        is_created = True
        obj = validate_and_create_object(model, **kwargs)
    return obj, is_updated, is_created



from .soft_deleted1 import *  # noqa
from .soft_deleted2 import *  # noqa
from .dated import *  # noqa
from .owned import *  # noqa
from .uided import *  # noqa
from .versioned import *  # noqa
from .historized import *  # noqa
from .base import *  # noqa



import factory
from django.utils.crypto import get_random_string

from ..models import MySimpleModel


class MySimpleModelFactory(factory.django.DjangoModelFactory):
    data = factory.LazyFunction(get_random_string)

    class Meta:
        model = MySimpleModel



from django.db import models

from pik.core.models import SoftDeleted


class BaseModel(models.Model):
    name = models.CharField(max_length=100, blank=True, null=True)


class NullRelatedModel(models.Model):
    nullable_base = models.ForeignKey(BaseModel, blank=True, null=True)


class BaseArchiveModel(SoftDeleted):
    name = models.CharField(max_length=100, blank=True, null=True)


class RelatedModel(models.Model):
    base = models.ForeignKey(BaseArchiveModel)
    set_null_base = models.ForeignKey(
        BaseArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_NULL)
    set_default_base = models.ForeignKey(
        BaseArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_DEFAULT)


class RelatedCousinModel(models.Model):
    related = models.ForeignKey(RelatedModel)
    set_null_related = models.ForeignKey(
        RelatedModel,
        blank=True, null=True, on_delete=models.deletion.SET_NULL)
    set_default_related = models.ForeignKey(
        RelatedModel,
        blank=True, null=True, on_delete=models.deletion.SET_DEFAULT)


class RelatedArchiveModel(SoftDeleted):
    base = models.ForeignKey(BaseArchiveModel)
    set_null_base = models.ForeignKey(
        BaseArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_NULL)
    set_default_base = models.ForeignKey(
        BaseArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_DEFAULT)


class RelatedCousinArchiveModel(SoftDeleted):
    related = models.ForeignKey(RelatedArchiveModel)
    set_null_related = models.ForeignKey(
        RelatedArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_NULL)
    set_default_related = models.ForeignKey(
        RelatedArchiveModel,
        blank=True, null=True, on_delete=models.deletion.SET_DEFAULT)



import pytest
from django.utils.crypto import get_random_string

from pik.core.shortcuts import get_object_or_none, validate_and_create_object, \
    validate_and_update_object, update_or_create_object
from .factories import MySimpleModelFactory
from ..models import MySimpleModel


@pytest.fixture(params=[
    (MySimpleModel, MySimpleModelFactory),
])
def test_model(request):
    return request.param


def test_get_object_or_none(test_model):
    model, factory = test_model
    objs = factory.create_batch(10)

    obj = get_object_or_none(model, data=get_random_string())
    assert obj is None

    obj = get_object_or_none(model, data=objs[-1].data)
    assert obj.pk


def test_validate_and_create_object(test_model):
    model, factory = test_model

    obj = validate_and_create_object(model, data=get_random_string())
    assert obj.pk


def test_validate_and_update_object__update(test_model):
    model, factory = test_model
    obj = factory.create()
    new_data = get_random_string()

    res_obj, is_updated = validate_and_update_object(obj, data=new_data)
    assert res_obj.pk == obj.pk
    assert is_updated
    assert res_obj.data == new_data


def test_validate_and_update_object__no_update(test_model):
    model, factory = test_model
    obj = factory.create()

    res_obj, is_updated = validate_and_update_object(obj, data=obj.data)
    assert res_obj.pk == obj.pk
    assert not is_updated
    assert res_obj.data == obj.data


def test_update_or_create_object__create_without_search(test_model):
    model, _ = test_model
    new_data = get_random_string()

    res_obj, is_updated, is_created = update_or_create_object(
        model, data=new_data)
    assert res_obj.pk
    assert not is_updated
    assert is_created


def test_update_or_create_object__create(test_model):
    model, factory = test_model
    obj = factory.create()
    new_data = get_random_string()

    res_obj, is_updated, is_created = update_or_create_object(
        model, search_keys=dict(data=new_data), data=new_data)
    assert res_obj.pk != obj.pk
    assert not is_updated
    assert is_created


def test_update_or_create_object__no_update(test_model):
    model, factory = test_model
    obj = factory.create()

    res_obj, is_updated, is_created = update_or_create_object(
        model, search_keys=dict(data=obj.data), data=obj.data)
    assert res_obj.pk == obj.pk
    assert not is_updated
    assert not is_created



from test_core_models import models


def test_built_in_cascade():
    """
    Verifies cascade deletion
    """
    base = models.BaseModel.objects.create(name='test')
    models.NullRelatedModel.objects.create(nullable_base=base)

    base.delete()
    assert not (models.NullRelatedModel.objects.exists())


def test_cascade_delete():
    """
    Verify that if we delete a model with the ArchiveMixin, then the
    delete cascades to its "parents", i.e. the models with foreign keys
    to it.
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    related = models.RelatedModel.objects.create(base=base)
    models.RelatedCousinModel.objects.create(related=related)
    related_archivable = models.RelatedArchiveModel.objects.create(
        base=base)
    cousin_archivable = models.RelatedCousinArchiveModel.objects.create(
        related=related_archivable)

    base.delete()

    assert not (models.RelatedModel.objects.exists())
    assert not (models.RelatedCousinModel.objects.exists())

    assert not (models.RelatedArchiveModel.objects.exists())
    assert (models.RelatedArchiveModel.all_objects.exists())
    related_archivable = models.RelatedArchiveModel.all_objects.get(
        pk=related_archivable.pk)
    assert (related_archivable.deleted) is not None

    assert not (models.RelatedCousinArchiveModel.objects.exists())
    assert (models.RelatedCousinArchiveModel.all_objects.exists())
    cousin_archivable = models.RelatedCousinArchiveModel.all_objects.get(
        pk=cousin_archivable.pk)
    assert (cousin_archivable.deleted) is not None


def test_cascade_delete_qs():
    """
    Verify that if we delete a model with the ArchiveMixin, then the
    delete cascades to its "parents", i.e. the models with foreign keys
    to it.
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    models.BaseArchiveModel.objects.create(name='test')
    models.BaseArchiveModel.objects.create(name='test')
    related = models.RelatedModel.objects.create(base=base)
    models.RelatedCousinModel.objects.create(related=related)
    related_archivable = models.RelatedArchiveModel.objects.create(
        base=base)
    models.RelatedCousinArchiveModel.objects.create(
        related=related_archivable)

    models.BaseArchiveModel.objects.all().delete()

    assert not (models.RelatedModel.objects.exists())
    assert not (models.RelatedCousinModel.objects.exists())
    assert not (models.RelatedArchiveModel.objects.exists())
    assert (models.RelatedArchiveModel.all_objects.exists())
    assert not (models.RelatedCousinArchiveModel.objects.exists())
    assert (models.RelatedCousinArchiveModel.all_objects.exists())


def test_cascade_nullable():
    """
    Verify that related models are deleted even if the relation is
    nullable.
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    base2 = models.BaseArchiveModel.objects.create(name='test2')
    related = models.RelatedModel.objects.create(
        base=base, set_null_base=base2, set_default_base=base2)
    archivable_related = models.RelatedArchiveModel.objects.create(
        base=base, set_null_base=base2, set_default_base=base2)
    models.RelatedCousinModel.objects.create(related=related)
    models.RelatedCousinArchiveModel.objects.create(
        related=archivable_related)

    base2.delete()

    assert 1 == models.BaseArchiveModel.objects.count()
    assert 1 == models.RelatedModel.objects.count()
    assert 1 == models.RelatedArchiveModel.objects.count()
    assert 1 == models.RelatedCousinModel.objects.count()
    assert 1 == models.RelatedCousinArchiveModel.objects.count()


def test_cascade_set_null():
    """
    Verify that related models are not deleted if on_delete is SET_NULL
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    base2 = models.BaseArchiveModel.objects.create(name='test2')
    related = models.RelatedModel.objects.create(
        base=base, set_null_base=base2)
    models.RelatedCousinModel.objects.create(related=related)

    base2.delete()

    assert 1 == models.BaseArchiveModel.objects.count()
    assert 1 == models.RelatedModel.objects.count()
    assert 1 == models.RelatedCousinModel.objects.count()

    assert (
        models.RelatedModel.objects.filter(pk=related.pk).exists())


def test_cascade_set_null_qs():
    """
    Verify that related models are not deleted if on_delete is SET_NULL
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    base2 = models.BaseArchiveModel.objects.create(name='test2')
    related = models.RelatedModel.objects.create(
        base=base, set_null_base=base2)
    models.RelatedCousinModel.objects.create(related=related)

    models.BaseArchiveModel.objects.filter(pk=base2.pk).delete()

    assert 1 == models.BaseArchiveModel.objects.count()
    assert 1 == models.RelatedModel.objects.count()
    assert 1 == models.RelatedCousinModel.objects.count()

    assert (
        models.RelatedModel.objects.filter(pk=related.pk).exists())


def test_cascade_set_default():
    """
    Verify that related models are not deleted if on_delete is SET_DEFAULT
    """
    base = models.BaseArchiveModel.objects.create(name='test')
    base2 = models.BaseArchiveModel.objects.create(name='test2')
    related = models.RelatedModel.objects.create(
        base=base, set_default_base=base2)
    models.RelatedCousinModel.objects.create(related=related)

    base2.delete()

    assert 1 == models.BaseArchiveModel.objects.count()
    assert 1 == models.RelatedModel.objects.count()
    assert 1 == models.RelatedCousinModel.objects.count()

    assert models.RelatedModel.objects.filter(pk=related.pk).exists()



from .dated import Dated
from .soft_deleted import SoftDeleted
from .owned import NullOwned, Owned
from .uided import Uided, PUided
from .versioned import Versioned
from .historized import Historized
from .base import BasePHistorical, BaseHistorical

__all__ = [
    'Dated', 'SoftDeleted', 'NullOwned', 'Owned', 'Uided', 'PUided',
    'Versioned', 'Historized', 'BasePHistorical', 'BaseHistorical']



import pytest

from .factories import MyUidedFactory, MyPUidedFactory
from ..models import MyUided, MyPUided


@pytest.fixture(params=[
    (MyUided, MyUidedFactory),
])
def uided_model(request):
    return request.param


@pytest.fixture(params=[
    (MyPUided, MyPUidedFactory),
])
def puided_model(request):
    return request.param


def test_uided_protocol(uided_model):
    model, factory = uided_model
    obj = factory.create()
    assert obj.uid
    assert obj.suid == str(obj.uid)
    assert obj.pk != obj.uid
    assert obj.stype == model._meta.model_name  # noqa


def test_puided_protocol(puided_model):
    model, factory = puided_model
    obj = factory.create()
    assert obj.uid
    assert obj.suid == str(obj.uid)
    assert obj.pk == obj.uid
    assert obj.stype == model._meta.model_name  # noqa



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _


class Owned(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_('User'),
        related_name='%(class)ss', db_index=True,
        on_delete=models.CASCADE, editable=False)

    class Meta:
        abstract = True


class NullOwned(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_('User'),
        related_name='%(class)ss', db_index=True,
        on_delete=models.CASCADE, editable=False,
        null=True)

    class Meta:
        abstract = True



import ulid
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.functional import cached_property


def _new_uid():
    return ulid.new().uuid


class Uided(models.Model):
    uid = models.UUIDField(unique=True, default=_new_uid, editable=False)

    @property
    def suid(self) -> str:
        return str(self.uid)

    @cached_property
    def stype(self):
        return ContentType.objects.get_for_model(type(self)).model

    def __str__(self):
        return self.suid

    class Meta:
        ordering = ['-uid']
        abstract = True


class PUided(models.Model):
    """
    Primary Uided
    """
    uid = models.UUIDField(primary_key=True, default=_new_uid, editable=False)

    @property
    def suid(self) -> str:
        return str(self.uid)

    @cached_property
    def stype(self):
        return ContentType.objects.get_for_model(type(self)).model

    def __str__(self):
        return self.suid

    class Meta:
        ordering = ['-uid']
        abstract = True



from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.crypto import get_random_string
from django.utils.module_loading import import_string


def create_user(username=None, password=None, **kwargs):
    factory_class = getattr(settings, 'USER_FACTORY_CLASS', '')
    try:
        factory = import_string(factory_class)
        user = factory.create(**kwargs)
        if password:
            user.set_password(password)
            user.save()
        return user
    except ImportError:
        pass

    User = get_user_model()  # noqa: pylint=invalid-name
    if username is None:
        username = kwargs.get(User.USERNAME_FIELD)

    if not username:
        username = get_random_string()

    kwargs.update({
        User.USERNAME_FIELD: username,
    })

    user = User.objects.create(**kwargs)
    if password:
        user.set_password(password)
        user.save()
    return user


def get_user(username=None):  # noqa: pylint=invalid-name
    User = get_user_model()  # noqa: pylint=invalid-name
    return User._default_manager.get_by_natural_key(username)  # noqa



from django.db import models

from pik.core.models import SoftDeleted


# Based on https://github.com/MnogoByte/django-permanent/blob/bdde297233eb7c83c862358854127c8654410aae/django_permanent/tests/test_app/models.py


class BaseTestModel(models.Model):
    class Meta:
        abstract = True

    def __str__(self):
        return str(self.pk)


class MyPermanentModel(SoftDeleted, BaseTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)
    pass


class RegularModel(BaseTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)
    pass


class RemovableRegularDepended(SoftDeleted, BaseTestModel):
    dependence = models.ForeignKey(RegularModel, on_delete=models.CASCADE)


class RemovableDepended(BaseTestModel):
    dependence = models.ForeignKey(MyPermanentModel, on_delete=models.CASCADE)


class NonRemovableDepended(SoftDeleted, BaseTestModel):
    dependence = models.ForeignKey(MyPermanentModel, on_delete=models.DO_NOTHING)


class NonRemovableNullableDepended(SoftDeleted, BaseTestModel):
    dependence = models.ForeignKey(MyPermanentModel, on_delete=models.SET_NULL, null=True)


class RemovableNullableDepended(SoftDeleted, BaseTestModel):
    dependence = models.ForeignKey(MyPermanentModel, on_delete=models.SET_NULL, null=True)


class PermanentDepended(SoftDeleted, BaseTestModel):
    dependence = models.ForeignKey(MyPermanentModel, on_delete=models.CASCADE)


class M2MFrom(BaseTestModel):
    pass


class PermanentM2MThrough(SoftDeleted):
    m2m_from = models.ForeignKey('M2MFrom', on_delete=models.CASCADE)
    m2m_to = models.ForeignKey('M2MTo', on_delete=models.CASCADE)


class M2MTo(BaseTestModel):
    m2m_from = models.ManyToManyField('M2MFrom', through=PermanentM2MThrough)



from collections import defaultdict

from django.contrib.admin.utils import NestedObjects
from django.core.exceptions import FieldDoesNotExist
from django.db import models, router, transaction
from django.db.models import Q, signals
from django.db.models.sql.where import WhereNode
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from ._collector_delete import Collector

assert Collector.delete


def _get_field_by_name(model, field):
    """
    Retrieve a field instance from a model by its name.
    """
    field_dict = {x.name: x for x in model._meta.get_fields()}  # noqa
    return field_dict[field]


def _has_field(model, field):
    try:
        model._meta.get_field(field)  # noqa
        return True
    except FieldDoesNotExist:
        return False


def _cascade_soft_delete(inst_or_qs, using, keep_parents=False):
    """
    Return collector instance that has marked ArchiveMixin instances for
    archive (i.e. update) instead of actual delete.
    Arguments:
        inst_or_qs (models.Model or models.QuerySet): the instance(s) that
            are to be deleted.
        using (db connection/router): the db to delete from.
        keep_parents (bool): defaults to False.  Determine if cascade is true.
    Returns:
        models.deletion.Collector: this is a standard Collector instance but
            the ArchiveMixin instances are in the fields for update list.
    """
    if not isinstance(inst_or_qs, models.QuerySet):
        instances = [inst_or_qs]
    else:
        instances = inst_or_qs

    deleted = now()

    # The collector will iteratively crawl the relationships and
    # create a list of models and instances that are connected to
    # this instance.
    collector = NestedObjects(using=using)
    collector.collect(instances, keep_parents=keep_parents)
    if collector.protected:
        raise models.ProtectedError("Delete protected", collector.protected)
    collector.sort()
    soft_delete_objs = collector.soft_delete_objs = defaultdict(set)

    for model, instances in list(collector.data.items()):
        # remove archive mixin models from the delete list and put
        # them in the update list.  If we do this, we can just call
        # the collector.delete method.
        if _has_field(model, 'deleted'):
            inst_list = [x for x in instances if x.deleted is None]
            deleted_on_field = _get_field_by_name(model, 'deleted')
            collector.add_field_update(deleted_on_field, deleted, inst_list)
            soft_delete_objs[model].update(inst_list)
            del collector.data[model]

    # If we use the NestedObjects collector instead models.deletion.Collector,
    # then the `collector.fast_deletes` will always be empty
    for i, q_set in enumerate(collector.fast_deletes):
        # make sure that we do archive on fast deletable models as
        # well.
        model = q_set.model
        if _has_field(model, 'deleted'):
            inst_list = [x for x in instances if x.deleted is None]
            deleted_on_field = _get_field_by_name(model, 'deleted')
            collector.add_field_update(deleted_on_field, deleted, inst_list)
            collector.fast_deletes[i] = q_set.none()

    return collector


def _delete_collected(collector):
    with transaction.atomic(using=collector.using, savepoint=False):
        result = collector.delete()
        for model, instances in collector.soft_delete_objs.items():
            if not model._meta.auto_created:  # noqa: pylint=protected-access
                for obj in instances:
                    signals.post_delete.send(
                        sender=model, instance=obj, using=collector.using
                    )
    return result


class _BaseSoftDeletedQuerySet(models.QuerySet):
    def delete(self):
        # doing an update is the most efficient, but does not promise
        # that the cascade will happen. E.g.
        # return self.update(deleted_on=timezone.now())

        # from django source
        # https://github.com/django/django/blob/1.8.6/django/db/models/query.py
        # Line: #L516
        assert self.query.can_filter(), \
            "Cannot use 'limit' or 'offset' with delete."

        # iterating and deleting ensures that the cascade delete will
        # occur for each instance.
        collector = _cascade_soft_delete(self.all(), self.db)
        self._result_cache = None
        return _delete_collected(collector)

    delete.alters_data = True  # type: ignore
    delete.queryset_only = True  # type: ignore

    def hard_delete(self):
        return models.QuerySet.delete(self)

    hard_delete.alters_data = True  # type: ignore
    hard_delete.queryset_only = True  # type: ignore

    def restore(self):
        return self.update(**{'deleted': None})

    restore.alters_data = True  # type: ignore
    restore.queryset_only = True  # type: ignore

    def get_restore_or_create(self, **kwargs):
        obj, created = self.model.all_objects.get_or_create(**kwargs)
        if not created and obj.deleted:
            obj.restore()
        return obj

    def update_restore_or_create(self, **kwargs):
        obj, created = self.model.all_objects.update_or_create(**kwargs)
        if not created and obj.deleted:
            obj.restore()
        return obj


class _AllWhereNode(WhereNode):
    pass


class _SoftDeletedObjectsWhereNode(WhereNode):
    pass


class _SoftObjectsWhereNode(WhereNode):
    pass


class SoftObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _SoftObjectsWhereNode
        self.query.add_q(Q(deleted=None))


class SoftDeletedObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _SoftDeletedObjectsWhereNode
        self.query.add_q(~Q(deleted=None))


class AllObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _AllWhereNode


class SoftDeleted(models.Model):
    """
    Soft deletable model. Inspired by:
    https://lucasroesler.com/2017/04/delete-or-not-to-delete/
    """
    deleted = models.DateTimeField(
        editable=False, null=True, blank=True, verbose_name=_('Deleted')
    )

    objects = SoftObjectsQuerySet.as_manager()
    deleted_objects = SoftDeletedObjectsQuerySet.as_manager()
    all_objects = AllObjectsQuerySet.as_manager()

    def delete(self, using=None, keep_parents=False):
        using = using or router.db_for_write(self.__class__, instance=self)

        assert self._get_pk_val() is not None, \
            "%s object can't be deleted because its %s attribute " \
            "is set to None." % (self._meta.object_name, self._meta.pk.attname)

        if self.deleted:
            return 0, {}  # short-circuit here to prevent lots of nesting

        if not self._meta.auto_created:
            signals.pre_delete.send(
                sender=self.__class__, instance=self, using=using)

        collector = _cascade_soft_delete(self, using, keep_parents)
        return _delete_collected(collector)

    delete.alters_data = True  # type: ignore

    def hard_delete(self, *args, **kwargs):
        return models.Model.delete(self, *args, **kwargs)

    def restore(self):
        self.deleted = None
        self.save(update_fields=['deleted'])

    class Meta:
        abstract = True



import pytest

from .factories import MyHistorizedFactory
from ..models import MyHistorized


@pytest.fixture(params=[
    (MyHistorized, MyHistorizedFactory),
])
def historized_model(request):
    return request.param


def test_historized_protocol(historized_model):
    model, factory = historized_model
    obj = factory.create()
    history = obj.history.all()
    assert obj.history
    assert obj.history.count() == 1

    obj.save()
    assert obj.history.count() == 2

    obj.delete()
    assert obj.history.count() == 0
    assert history.count() == 3

    hist_obj1 = history.last()
    hist_obj2 = history[1]
    hist_obj3 = history.first()
    assert isinstance(hist_obj3.history_id, int)
    assert hist_obj3.pk == hist_obj3.history_id
    assert hist_obj3.history_change_reason is None
    assert hist_obj3.history_id > hist_obj1.history_id

    assert hist_obj1.history_type == '+'
    assert hist_obj2.history_type == '~'
    assert hist_obj3.history_type == '-'



import datetime

import pytest
from freezegun import freeze_time

from ..models import MyDated
from .factories import MyDatedFactory


@pytest.fixture(params=[
    (MyDated, MyDatedFactory),
])
def dated_model(request):
    return request.param


@freeze_time("2012-01-14 03:21:34")
def test_dated_protocol(dated_model):
    model, factory = dated_model

    obj = factory.create()
    assert obj.created == datetime.datetime(2012, 1, 14, 3, 21, 34)
    assert obj.updated == datetime.datetime(2012, 1, 14, 3, 21, 34)

    with freeze_time("2012-01-14 03:21:55"):
        obj.save()
    assert obj.created == datetime.datetime(2012, 1, 14, 3, 21, 34)
    assert obj.updated == datetime.datetime(2012, 1, 14, 3, 21, 55)
